local addonName, addonTable = ...

-- Debugging to tekDebug

local debug = false
--[===[@debug@
debug = true
--@end-debug@]===]

local name, _, _, enabled = GetAddOnInfo("tekDebug")
local player = UnitName("player")
local enabled = GetAddOnEnableState(player, "tekDebug") == 2
if enabled and not IsAddOnLoaded("tekDebug") then
  local succ, err = LoadAddOn(name)
end
local debugf = debug and tekDebug and tekDebug:GetFrame(addonName)
local function DebugPrint(...) if debugf then debugf:AddMessage(string.join(", ", tostringall(...))) end end

local strfind, strformat, strgsub, strlower, strmatch, tostring = string.find, string.format, string.gsub, string.lower, strmatch, tostring
local _, date, ipairs, next, pairs, select, time, tinsert, tsort, type, wipe = _, date, ipairs, next, pairs, select, time, table.insert, table.sort, type, wipe

-- LibDataBroker
local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
local LDBIcon = ldb and LibStub("LibDBIcon-1.0",true)
local LibQTip = LibStub('LibQTip-1.0')
local BF = LibStub("AceAddon-3.0"):NewAddon(addonName, "AceConsole-3.0", "AceEvent-3.0")

if debug then
  Broker_Followers = BF
end

BF.a = addonTable
local L = LibStub("AceLocale-3.0"):GetLocale(addonName, false)
local DB_VERSION = 1
BF.a.followerList = {}
BF.a.collectedFollowersList = {}
BF.a.reverseCounterLookup = {}
BF.a.threatStrings = {
  [1] = "Wild Aggression",
  [2] = "Massive Strike",
  [3] = "Group Damage",
  [4] = "Magic Debuff",
  [6] = "Danger Zones",
  [7] = "Minion Swarms",
  [8] = "Powerful Spell",
  [9] = "Deadly Minions",
  [10] = "Timed Battle",
  [11] = "Orc",
  [12] = "Ogre",
  [13] = "Undead",
  [14] = "Arakkoa",
  [15] = "Beast",
  [16] = "Demon",
  [17] = "Breaker",
  [18] = "Fury",
  [19] = "Aberration",
  [20] = "Primal",
  [21] = "Mountains",
  [22] = "Desert",
  [23] = "Snow",
  [24] = "Underground",
  [25] = "Jungle",
  [26] = "Forest",
  [27] = "Town",
  [28] = "Swamp",
  [29] = "Plains",
}
--[[====
Threats
6:  Danger Zones      Environments            Races          
9:  Deadly Minions    22: Desert              19: Aberration 
3:  Group Damage      26: Forest              14: Arakkoa    
4:  Magic Debuff      58: Hearthstone Table   15: Beast      
2:  Massive Strike    25: Jungle              17: Breaker    
7:  Minion Swarms     21: Mountains           16: Demon      
8:  Powerful Spell    29: Plains              18: Fury       
10: Timed Battle      23: Snow                12: Ogre       
1:  Wild Aggression   28: Swamp               11: Orc        
                      27: Town                20: Primal     
                      24: Underground         13: Undead     
====]]--


local defaults = {
  profile = {
    version = 1,
    minimapIcon = {
      hide = true,
      minimapPos = 220,
      radius = 80,
    },
  },
  global = { 
    version = 1,
  },
}

local rarityColors = {}
function BF:OnInitialize() -- ADDON_LOADED, only saved variables are guaranteed to be present
  self.db = LibStub("AceDB-3.0"):New("Broker_Follower_DB", defaults, true)
  self:MigrateSavedData()

  LibStub("AceConfig-3.0"):RegisterOptionsTable(addonName, self.GenerateOptions)
  self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName)

  self.BFLDB = ldb:NewDataObject(addonName, {
    type = "data source",
    text = "",
    icon = "Interface\\Icons\\inv_misc_lockboxghostiron",
    OnClick = function(clickedframe, button)
      if button == "RightButton" then 
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame) 
      end
    end,
    OnEnter = function(tt)
      local tooltip = LibQTip:Acquire("Broker_Followers_Tooltip", 1, "LEFT") 
      self.tooltip = tooltip
      self:PrepareTooltip(tooltip, not IsShiftKeyDown()) 
      tooltip:SmartAnchorTo(tt)
      tooltip:Show()
    end,
    OnLeave = function(tt)
      LibQTip:Release(BF.tooltip)
      self.tooltip = nil
    end,
  })

  LDBIcon:Register(addonName, self.BFLDB, self.db.profile.minimapIcon)
  BF:UpdateFollowerCounters()
  for i=0,7 do
    rarityColors[i] = "|c"..select(4,GetItemQualityColor(i)).."%s|r"
  end
end

-- Options Handling

function BF:GenerateOptions()
  if not BF.options then

    local function newCounter()
      local i = 0
      return function( inc )
          i = i + 1
          return i
        end
    end
    local count = newCounter()

    BF.options = {
      name = addonName,
      type = 'group',
      handler = BF,
      args = {
        genconfig = {
          name = L["General Settings"],
          type = 'group',
          order = 1,
          get = function(info) return PT.db.profile[info[#info]] end,
          set = function(info, value) PT.db.profile[info[#info]] = value end,
          args = {
            minimapIcon = {
              type = "toggle",
              name = L["Show minimap button"],
              order = count(),
              get = function(info) return not BF.db.profile.minimapIcon.hide end,
              set = function(info, value)
                BF.db.profile.minimapIcon.hide = not value
                if value then LDBIcon:Show(addonName) else LDBIcon:Hide(addonName) end
              end,
            },
          },
        },
      },
    }
  end 
  return BF.options
end

function BF:PrepareTooltip(tooltip, tableInsteadOfList)
  UIParentLoadAddOn("Blizzard_GarrisonUI")
  
  if not next(BF.a.collectedFollowersList) then
    BF:UpdateFollowerCounters()
  end
  
  local function sortFollowers(f1, f2)
    local f1a, f2a = BF.a.collectedFollowersList[f1.followerID], BF.a.collectedFollowersList[f2.followerID]
    if f1.isCollected and f2.isCollected  then
      for i=1,10 do
        if f1a.counters[i] and not f2a.counters[i] then
          return true
        end
        if not f1a.counters[i] and f2a.counters[i] then
          return flase
        end
      end
      return f1.name > f2.name
    else
      if not f1.isCollected and not f2.isCollected then
        return f1.name > f2.name
      else
        return f1.isCollected
      end
    end
  end
  
  local function sortAbilities(a1, a2)
    if not a1.isTrait and a2.isTrait  then
      return true
    end
    if a1.isTrait and  not a2.isTrait  then
      return false
    end
    local a1c = next(a1.counters)
    local a2c = next(a2.counters)
    if a1.isTrait and a2.isTrait  then
      if a1c or a2c then
        return (a1c or 0) > (a2c or 0)
      else
        return a1.name > a2.name
      end
    end
    a1c = next(a1.counters)
    a2c = next(a2.counters)
    return a1c < a2c
  end   
  local Columns = {}
  local Rows = {}
  tooltip.Columns = Columns
  tooltip.Rows = Rows
  local function MyAddColumn(nameType, displayName)
    if nameType then
      if not displayName then
        BF:Debug("MyAddColumn: missing displayName for", nameType)
        displayName = nameType
      end
      if not formatFunc then
        --formatFunc = CellFormat.Default
      end
      
      if not Columns[nameType] then
        local colindex = tooltip:AddColumn("CENTER")
        tooltip:SetCell(1, colindex, displayName)
        
        Columns[nameType] = { col = colindex }
      else
        BF:Debug("a column already exists for", nameType)
      end
    end
  end
  local function AddSpanner(name, abbr, first, last)
    -- TODO - can we, from the size of span and maybe the character count of
    -- name make a guess as to whether we could use the name instead of Abbr?
    if first <= last then
      tooltip:SetCell(1, first, abbr, nil, nil, 1 + last - first)
    end
  end
  colors = {
    green = {r=0, g=1, b=0, a=1},
    white = {r=1, g=1, b=1, a=1},
    red = {r=1, g=0, b=0, a=1},
    
    lightGray = {r=(108/255), g=(115/255), b=(120/255), a=1},
    green = {r=(32/255), g=(255/255), b=(32/255), a=1},
    red = {r=(255/255), g=(25/255), b=(25/255), a=1},
    yellow = {r=(255/255), g=(228/255), b=(0/255), a=1},
    
    --lightGray = {r=0.25, g=0.25, b=0.25, a=1},
    darkGray = {r=0.1, g=0.1, b=0.1, a=1},
    lineGrey = {r=(108/255), g=(115/255), b=(120/255), a=1},
    --lightGray = {r=(82/255), g=(91/255), b=(97/255), a=1},
  }
  local function AddSeparator(tooltip)
    tooltip:AddSeparator(1, colors.lineGrey.r, colors.lineGrey.g, colors.lineGrey.b, colors.lineGrey.a)
  end
  
  local globaldb = self.db.global
  local profiledb = self.db.profile
  
  if tableInsteadOfList then
    -- 9x9 table, each cell showing the number of follower who counters both
    -- threats from column and row
    local mechanics = {}
    local q = C_Garrison.GetFollowerAbilityCounterMechanicInfo
    for _, aid in pairs({11, 100, 168, 148, 160, 101, 105, 157, 122}) do
      local mid, _, tex = q(aid)
      mechanics[mid], mechanics[tex:lower():gsub("%.blp$","")] = aid, aid
    end
    
    local function getMechanicInfo(mid)
      if populateMechanics then
        populateMechanics = populateMechanics()
      end
      if mechanics[mid] then
        return C_Garrison.GetFollowerAbilityCounterMechanicInfo(mechanics[mid])
      end
    end
    
    tooltip:AddHeader()
    tooltip:AddHeader()
    MyAddColumn("icon","")
    tooltip:SetCell( 2,1,"")
    for k=1,10 do
      local _, _, tex = getMechanicInfo(k)
      if tex then
        tex = string.format("|T%s:0|t", tex)
        local row = tooltip:AddHeader()
        --AddSeparator(tooltip)
        MyAddColumn(tostring(k), tex)
        tooltip:SetCell( row, 1, tex)
        Rows[k] = row
      end
    end 
    local counterSum = {}
    tooltip.counterDualSum = {}
    local function addDualCounter( c1, c2 )
      tooltip.counterDualSum[c1] = tooltip.counterDualSum[c1] or {}
      tooltip.counterDualSum[c2] = tooltip.counterDualSum[c2] or {}
      tooltip.counterDualSum[c2][c1] = (tooltip.counterDualSum[c2][c1] or 0) + 1
      tooltip.counterDualSum[c1][c2] = (tooltip.counterDualSum[c1][c2] or 0) + 1
    end
    local allFollowers = C_Garrison.GetFollowers()
    if #allFollowers == 0 then
      return 
    end
    table.sort(allFollowers, sortFollowers)
    for _, followerData in pairs(  allFollowers ) do
      if followerData.isCollected and ( followerData.status ~= "Inactive" ) then
        local abilities = C_Garrison.GetFollowerAbilities(followerData.followerID)
        local counters = {}
        for id, abilityData in pairs( abilities ) do
          
          local counterID, _ = next(abilityData.counters)
          if not abilityData.isTrait then
            table.insert( counters, counterID)
            counterSum[counterID] = ( counterSum[counterID] or 0 ) + 1
            tooltip:SetCell(2, Columns[tostring(counterID)].col, counterSum[counterID])
          end
        end
        if #counters > 1 then
          addDualCounter( counters[1], counters[2] )
        end
      end
    end
    local r,c
    for r, cData in pairs(tooltip.counterDualSum) do
      for c, num in pairs(cData) do
        tooltip:SetCell( Rows[r], Columns[tostring(c)].col, num)
      end
    end
  else
    -- Tooltip: List of followers with 5 columns for counters/traits
    tooltip:AddHeader()
    MyAddColumn("counter1","")
    MyAddColumn("counter2","")
    AddSpanner("Threats", "vs", 2, 3)
    MyAddColumn("trait1","")
    MyAddColumn("trait2","")
    MyAddColumn("trait3","")
    AddSpanner("Traits", "Traits", 4, 6)
    local allFollowers = C_Garrison.GetFollowers()
    if #allFollowers == 0 then
      return 
    end
    table.sort(allFollowers, sortFollowers)
    for _, followerData in pairs(  allFollowers ) do
      if followerData.isCollected then
        local row = tooltip:AddLine("")
        local coloredName = string.format(rarityColors[followerData.quality], followerData.name)
        tooltip:SetCell(row, 1, string.format("%s (%d)%s", coloredName, followerData.level == 100 and followerData.iLevel or followerData.level, (followerData.status == "Inactive" and " X" or "") ) )
        local counter, trait = 1,1
        local abilities = C_Garrison.GetFollowerAbilities(followerData.followerID)
        table.sort(abilities, sortAbilities)
        
        for id, abilityData in pairs( abilities ) do
          local _, counterData = next(abilityData.counters)
          
          if  abilityData.isTrait then
            tooltip:SetCell(row, Columns["trait"..trait].col, string.format("|T%s:0|t",counterData and counterData.icon or abilityData.icon))
            trait = trait+1
          else
            tooltip:SetCell(row, Columns["counter"..counter].col, string.format("|T%s:0|t",counterData and counterData.icon or abilityData.icon) )
            counter = counter+1
          end
        end
      end
    end
  end
end

-- 
function BF:OnEnable()
  self:RegisterEvent("GARRISON_FOLLOWER_XP_CHANGED",  "UpdateFollowerCounters")
  self:RegisterEvent("GARRISON_FOLLOWER_ADDED",       "UpdateFollowerCounters")
  self:RegisterEvent("GARRISON_FOLLOWER_LIST_UPDATE", "UpdateFollowerCounters")
  self:RegisterEvent("GARRISON_FOLLOWER_REMOVED",     "UpdateFollowerCounters")
  self:RegisterEvent("GARRISON_FOLLOWER_XP_CHANGED",  "UpdateFollowerCounters")
end

function BF:OnDisable()
  self:RegisterEvent("GARRISON_FOLLOWER_XP_CHANGED" )
  self:RegisterEvent("GARRISON_FOLLOWER_ADDED"      )
  self:RegisterEvent("GARRISON_FOLLOWER_LIST_UPDATE")
  self:RegisterEvent("GARRISON_FOLLOWER_REMOVED"    )
  self:RegisterEvent("GARRISON_FOLLOWER_XP_CHANGED" )
end

---- Parse Follower List

function BF:UpdateFollowerCounters()
  local followerData, counterID, followerAbilitiesData
  BF.a.followerList = C_Garrison.GetFollowers()
  wipe( BF.a.collectedFollowersList )
  wipe( BF.a.reverseCounterLookup )
  BF.a.collectedFollowersList = {}
  BF.a.reverseCounterLookup = {}

  for _, followerData in ipairs( BF.a.followerList ) do

    if followerData.isCollected then
      BF:Debug( strformat("Found %s (%s) L:%d Q:%d", followerData.name, followerData.followerID, followerData.level or -1, followerData.quality or -1) )
      followerData.counters = {}

      for _, followerAbilitiesData in ipairs(C_Garrison.GetFollowerAbilities(followerData.followerID)) do
        counterID, _ = next( followerAbilitiesData.counters)

        if counterID then
          BF:Debug("Counters "..followerAbilitiesData.counters[counterID].name)
          
          followerData.counters[counterID] = true

          BF.a.reverseCounterLookup[counterID] = BF.a.reverseCounterLookup[counterID] or {}
          tinsert(BF.a.reverseCounterLookup[counterID], followerData.followerID)
        end
      end
      BF.a.collectedFollowersList[followerData.followerID] =  followerData
    end
  end
end

---- Database Maintenance

function BF:MigrateSavedData()
  local  globaldb = self.db.global
  local profiledb = self.db.profile

  -- all db sections are current we can exit immediately
  if profiledb.version == DB_VERSION and
      globaldb.version == DB_VERSION then
    return
  else
    BF:Print(L["DB_VERSION_UPGRADE_PURGE"])
  end
  
  -- Migrating from before addon began keeping track of database changes
  --[[
  if globaldb.version < X then
    -- Upgrade from before version X
  end
  if profiledb.version < X then
    -- Upgrade from before version X
  end
  --]]--

  profiledb.version = DB_VERSION
  globaldb.version = DB_VERSION
end

function BF:Debug(...)
  DebugPrint( ... )
end